 /************************************************************************\
 * Fork Bomb Defuser 0.1 (c)1999 Rohit Singh <rohit_singh@hotmail.com>    *
 * 15th Dec'99	                                                          *
 *                                                                        *	
 * Permission to use, copy, modify, and distribute this software and its  *
 * documentation for any purpose and without fee is hereby granted,       *
 * provided that the above copyright notice appear in all copies and that *
 * both that copyright notice and this permission notice appear in        *
 * supporting documentation.                                              *
 *                                                                        *
 * This file is provided AS IS with no warranties of any kind. The author *
 * shall have no liability with respect to the infringement of copyrights *
 * trade secrets or any patents by this file or any part thereof.  In no  *
 * event will the author be liable for any lost revenue or profits or     *
 * other special, indirect and consequential damages.                     *
 * 									  *		
 * DISCLAIMER : This program is in its development stage, and because of  *
 * its nature it may crash your system. USE IT AT YOUR OWN RISK. The      *
 * author shall NOT be responsible for any damage or harm to your syatem  *
 * or data.                                                               *
 *                                                                        *
 * Please send all bug-reports, comments, Patches(!) and suggestions to:  *
 * <rohit_singh@hotmail.com>                                              *
 * If you find this program useful, please send me an email so that I can *
 * tell you when an updated version is available.                         *
 \************************************************************************/  
                                                                        
 /***********************************************************************\  
 * Fork Bomb Defuser (rexFBD) is a Linux Kernel Module which detects     *
 * and disables "fork bombs". Fork Bombs are programs/shell scripts      *
 * which (either intentionally or accidently) eat up process space       *
 * by creating new processes/shells repeatedly using the fork() system   *
 * call. Once a fork bomb bursts, it can be very difficult to detect     *
 * and disable it as the system comes to a grinding halt within no time. *
 *									 *
 * Although, the maximum number of processes which can be forked by a    *
 * user can be predefined using configuration files in Linux, there is   *
 * no mechanism (as yet) to control the "rate of Forking". This is       *
 * where rexFBD can help...           	                                 *
 * You can pass "max_tasks_per_user" and "max_forks_per_second"          *
 * to the module as parameters, while insmod'ing it.                     *
 * rexFBD also LOGS all possible Fork Bombs in /var/log/messages in the  *
 * following manner:	                                                 *
 * Dec 12 16:32:46 localhost kernel: Possible Fork Bomb by uid(0) Defused*
 \***********************************************************************/  
 
/*                                                                     
 * INSTALLATION:	
 * $ make 
 * $ su
 * # /sbin/insmod rexFBD.o [max_forks_per_second=XX] [max_tasks_per_user=NN]
 * where XX and NN are the respective values for the parameters. Both
 * parameters are OPTIONAL and their default values are 50 and 100
 * respectively.
 *
 * Remove the module with 
 * # /sbin/rmmod rexFBD
 *
 */


#include <linux/kernel.h>
#include <linux/module.h>

#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif


#include <sys/syscall.h>
#include <linux/sched.h>

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
#endif /*KERNEL_VERSION*/

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>		
#endif

extern void *sys_call_table[];
  
int 	max_forks_per_second=50;
int 	max_tasks_per_user=100;

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,2,0)
MODULE_PARM(max_forks_per_second, "i");                         
MODULE_PARM(max_tasks_per_user, "i");                         
#endif

int (*orig_sys_fork) (struct pt_regs);

int modified_sys_fork(struct pt_regs regs )
{
	unsigned long 	current_time=0;

	int 	forks_in_last_second = 0;
	int 	tasks_by_this_user=0;
	int 	invoking_uid;

	struct task_struct *p;

	current_time=jiffies;

	forks_in_last_second=0;
	invoking_uid=current->uid;

	read_lock(&tasklist_lock);
	for_each_task(p)
	{
		if ((p->uid)==invoking_uid)
		{
			tasks_by_this_user++;
			if ( (current_time - (p->start_time)) <= 100)
				forks_in_last_second++;
		}
	}
	read_unlock(&tasklist_lock);

	if ( (tasks_by_this_user   >= max_tasks_per_user) 
	   ||(forks_in_last_second >= max_forks_per_second))
	{
		printk(KERN_INFO "Possible Fork Bomb by uid(%d) Defused\n", invoking_uid);
		forks_in_last_second=0;
		tasks_by_this_user=0;
		return(-EAGAIN); 
	}
	
	else 
	{
	return (orig_sys_fork(regs));
	}
}


int init_module()
{
	orig_sys_fork = sys_call_table[SYS_fork];
	sys_call_table[SYS_fork] = modified_sys_fork;
	printk(KERN_INFO "Fork Bomb Defuser by Rohit Singh loaded as kernel module\n");

	return(0);
}


void cleanup_module()
{
	if (sys_call_table[SYS_fork] != modified_sys_fork)
		printk("Fork Bomb Defuser on fire!\n");
	sys_call_table[SYS_fork] = orig_sys_fork;
	printk("Fork Bomb Defuser unloaded from kernel\n");
}

